import collections
from src import data_generator


def test_generate_no_constraint():
    data_type = 'integer'
    rows_to_gen = 20
    avg_width = None
    table_name = 'atp_players'
    min_bound = 1
    max_bound = 21
    numeric_precision = 32

    values = data_generator.generate_no_constraint(data_type=data_type,
                                                   rows_to_gen=rows_to_gen,
                                                   avg_width=avg_width,
                                                   table_name=table_name,
                                                   numeric_precision=numeric_precision,
                                                   min_bound=min_bound,
                                                   max_bound=max_bound)

    max_value = max(values)
    min_value = min(values)

    assert values
    assert len(values) == 20
    assert min_value >= 1
    assert max_value <= 21


def test_generate_primary_key_number():

    data_type = 'integer'
    rows_to_gen = 30
    avg_width = None
    table_name = 'atp_players'
    min_bound = 1
    max_bound = 40

    values = data_generator.generate_pk(data_type=data_type,
                                        rows_to_gen=rows_to_gen,
                                        avg_width=avg_width,
                                        table_name=table_name,
                                        min_value=min_bound,
                                        max_value=max_bound,)

    max_value = max(values)
    min_value = min(values)
    duplicate = [item for item, count in collections.Counter(
        values).items() if count > 1]

    assert values
    assert len(values) == 30
    assert min_value >= 1
    assert max_value <= 40
    # no duplicated values expected for primary key
    assert len(duplicate) is 0


def test_generate_pk_mf():
    # Amount of values to generate is greater than range between min_bound and max_bound

    data_type = 'integer'
    rows_to_gen = 800
    avg_width = None
    table_name = 'atp_players'
    min_bound = 1
    max_bound = 40

    values = data_generator.generate_pk(data_type=data_type,
                                        rows_to_gen=rows_to_gen,
                                        avg_width=avg_width,
                                        table_name=table_name,
                                        min_value=min_bound,
                                        max_value=max_bound)

    max_value = max(values)
    min_value = min(values)

    assert values
    assert len(values) == 800
    assert min_value >= 1
    assert max_value >= 40
    assert max_value <= 800


def test_generate_foreign_key():
    data_type = 'integer'
    rows_to_gen = 30
    avg_width = None
    table_name = 'atp_rankings'
    column_name = 'loser_id'
    min_bound = 1
    max_bound = 78
    column_information = {'loser_id': {}}

    ref_values = {1, 2, 3, 5, 7, 9, 11, 13, 14, 15, 16, 17, 18, 29, 30, 35, 36, 37, 38, 39, 40,
                  41, 42, 43, 44, 54, 55, 66, 68, 69, 70, 71, 72, 73, 74, 75, 77, 78, 79, 81, 80}

    pk_values = dict()
    pk_values['atp_players'] = dict()
    pk_values['atp_players']['player_id'] = ref_values

    fk_ref_info = dict()
    fk_ref_info['loser_id'] = dict()
    fk_ref_info['loser_id']['ref_column'] = 'player_id'
    fk_ref_info['loser_id']['ref_table'] = 'atp_players'

    values = data_generator.generate_fk(data_type=data_type,
                                        rows_to_gen=rows_to_gen,
                                        table_name=table_name,
                                        column_name=column_name,
                                        pk_values=pk_values,
                                        fk_ref_info=fk_ref_info,
                                        min_bound=min_bound,
                                        max_bound=max_bound)

    max_value = max(values)
    min_value = min(values)
    ref_available = True
    for fk in values:
        if fk not in pk_values['atp_players'].get('player_id'):

            ref_available = False
        continue

    assert values
    assert len(values) == 30
    assert min_value >= 1
    assert max_value <= 78
    # foreign key has to be available in provided pk_values
    assert ref_available is True


def test_generate_fk_extended():
    # amount of values to generate is higher than amount of possible values between boundaries
    data_type = 'integer'
    rows_to_gen = 40
    table_name = 'atp_rankings'
    column_name = 'loser_id'
    min_bound = 2
    max_bound = 8

    ref_values = {1, 2, 3, 5, 7, 9, 11, 13, 14, 15, 16, 17, 18, 29, 30, 35, 36, 37, 38, 39, 40,
                  41, 42, 43, 44, 54, 55, 66, 68, 69, 70, 71, 72, 73, 74, 75, 77, 78, 79, 81, 80}

    pk_values = dict()
    pk_values['atp_players'] = {}
    pk_values['atp_players']['player_id'] = ref_values

    fk_ref_info = dict()
    fk_ref_info['loser_id'] = dict()
    fk_ref_info['loser_id']['ref_column'] = 'player_id'
    fk_ref_info['loser_id']['ref_table'] = 'atp_players'

    values = data_generator.generate_fk(data_type=data_type,
                                        rows_to_gen=rows_to_gen,
                                        table_name=table_name,
                                        column_name=column_name,
                                        pk_values=pk_values,
                                        fk_ref_info=fk_ref_info,
                                        min_bound=min_bound,
                                        max_bound=max_bound)

    max_value = max(values)
    min_value = min(values)

    assert values
    assert len(values) == 40
    assert min_value >= 1
    assert max_value <= 80



from typing import Dict, Any, List, Optional, Union
import random
from src.generator import generator_utils
import datetime
from src import postgres


def create_insert_query(total_rows: int, insert_query: str, table_name: str,
                        column_names: List[str],
                        table_information, column_information: Dict[str, Dict[str, Any]],
                        insert_dict: Dict) -> Dict:
    # Numerate for columns without frequency
    i = 0
    print(f'Creating insert query for the "{table_name}" table...')

    for _ in range(round(total_rows)):
        column_values = list()
        insert_query += "INSERT INTO {table_name}("
        insert_query += '{0}{1}'.format(', '.join(column_names), ') VALUES (')
        random_frac = random.random()

        for column_info in table_information[table_name]["column_information"].values():
            column_name = column_info.get("column_name")
            data_type = column_info.get("data_type")

            if data_type not in postgres.DataTypes.SUPPORTED_TYPES:
                data_type = 'not_supported'

            max_length = column_info.get("max_length")
            numeric_precision = column_info.get("numeric_precision")
            numeric_precision_radix = column_info.get(
                "numeric_precision_radix")
            numeric_scale = column_info.get("numeric_scale")

            """ The next line messes up the primary key insertion. It forces to diregard 
                the generated_vals and does an auto increment where it is specified as a 
                column_default. 

                This can subsequently lead to insertion errors on foreign keys since primary 
                keys might not be the same in generated database and the ones used to 
                generate foreign key values in the tool. 

            """
            # if column_info.get("column_default"):
            # continue

            generated_vals = None
            generated_freqs: Optional[Union[List[float], bool]] = None
            null_frac = None
            column_stats = None

            if column_name in table_information[table_name]["pg_stats"]:
                column_stats = table_information[table_name]["pg_stats"][column_name]

            if column_stats:
                if "generated_vals" in column_information[column_name] \
                        and column_information[column_name]["generated_vals"]:
                    generated_vals = column_information[column_name]["generated_vals"]
                    generated_freqs = column_information[column_name]["generated_freqs"]
                    null_frac = column_stats["null_frac"]

            def numeric():
                # Columns containing MCV
                if generated_vals and generated_freqs:
                    if null_frac and random_frac <= null_frac:
                        column_values.append("{0}".format('null'))
                    else:
                        column_values.append("{0}".format(
                            generator_utils.random_choices(generated_vals, generated_freqs)))
                # All distinct columns
                elif generated_vals:
                    column_values.append(
                        "{0}".format(generated_vals[i]))
                else:
                    column_values.append(
                        "{0}".format(generator_utils.random_number(numeric_precision,
                                                                   numeric_precision_radix,
                                                                   numeric_scale)))

            def date():
                if generated_vals and generated_freqs:
                    if null_frac and random_frac <= null_frac:
                        column_values.append("{0}".format('NULL'))
                    else:
                        column_values.append("'{0}'".format(
                            generator_utils.random_choices(generated_vals, generated_freqs)))
                # All distinct columns
                elif generated_vals:
                    column_values.append("'{0}'".format(generated_vals[i]))
                else:
                    START_DATE = datetime.date(year=1950, month=1, day=1)
                    END_DATE = datetime.date.today()
                    column_values.append("'{0}'".format(
                        generator_utils.random_date(START_DATE, END_DATE)))

            def boolean():
                column_values.append("{0}".format(
                    generator_utils.random_boolean()))

            def varchar():
                if generated_vals and generated_freqs:
                    if null_frac and random_frac <= null_frac:
                        column_values.append("{0}".format('NULL'))
                    else:
                        column_values.append("'{0}'".format(
                            generator_utils.random_choices(generated_vals, generated_freqs)))
                # All distinct columns
                elif generated_vals:
                    column_values.append("'{0}'".format(generated_vals[i]))
                else:
                    column_values.append("'{0}'".format(
                        generator_utils.random_word(max_length / 2.5)))

            def geometry():
                # Columns containing MCV
                if generated_vals and generated_freqs:
                    if null_frac and random_frac <= null_frac:
                        column_values.append("{0}".format('NULL'))
                    else:
                        column_values.append("'{0}'".format(
                            generator_utils.random_choices(generated_vals, generated_freqs)))
                # All distinct columns
                elif generated_vals:

                    column_values.append(
                        "{0}".format(generated_vals[i]))
                else:
                    column_values.append("{0}".format('NULL'))

            def bytea():
                # Columns containing MCV
                if generated_vals and generated_freqs:
                    if null_frac and random_frac <= null_frac:
                        column_values.append("{0}".format('NULL'))
                    else:
                        column_values.append("'{0}'".format(
                            generator_utils.random_choices(generated_vals, generated_freqs)))
                # All distinct columns
                elif generated_vals:

                    column_values.append(
                        "'{0}'".format(generated_vals[i]))
                else:
                    column_values.append("{0}".format('NULL'))

            def no_type():
                """ print(
                    f'The "{data_type}" data type is not supported. '
                    f'Skipping the table\'s "{table_name}" data generation...') """
                column_values.append("{0}".format('NULL'))

            funcDict = {'text': varchar,
                        'character varying': varchar,
                        'character': varchar,
                        'smallint': numeric,
                        'integer': numeric,
                        'bigint': numeric,
                        'decimal': numeric,
                        'numeric': numeric,
                        'double precision': numeric,
                        'real': numeric,
                        'date': date,
                        'timestamp': date,
                        'timestamp without time zone': date,
                        'boolean': boolean,
                        'bool': boolean,
                        'point': geometry,
                        'line': geometry,
                        'polygon': geometry,
                        'circle': geometry,
                        'box': geometry,
                        'lseg': geometry,
                        'path': geometry,
                        'bytea': bytea,
                        'not_supported': no_type}

            funcDict[data_type]()

        i = i + 1
        insert_query += '{0}{1}'.format(', '.join(column_values), ');')

    insert_dict[table_name] = insert_query

    return insert_dict



from collections import defaultdict
import datetime
from src import postgres
import sys
from psycopg2._psycopg import cursor
from typing import Any, List, Optional, Tuple, Dict


def get_date_bound(most_common_values: List[str], bound: str) -> datetime.date:

    most_common_dates = [old_format.replace(
        "-", '') for old_format in most_common_values]

    if bound == "start":
        date = min(most_common_dates)
    elif bound == "end":
        date = max(most_common_dates)

    if date is not None:
        date_y = int(date[:4])
        date_m = int(date[4:6])
        date_d = int(date[6:8])

        if bound == "start":
            START_DATE = datetime.date(year=date_y, month=date_m, day=date_d)
            return START_DATE

        elif bound == "end":
            END_DATE = datetime.date(year=date_y, month=date_m, day=date_d)
            return END_DATE

    else:
        if bound == "start":
            START_DATE = datetime.date(year=1950, month=1, day=1)
            return START_DATE
        elif bound == "end":
            END_DATE = datetime.date.today()
            return END_DATE


def days_between(date1: Any, date2: Any) -> int:
    return (date2 - date1).days


def determine_tables_to_generate(table_results, u_tables) -> List[str]:
    # Save all database tables in list db_tables
    db_tables = list()
    for table_entry in table_results:
        db_tables.append(table_entry[1])

    # Arrange user specified table into user_tables
    user_tables = list()

    if u_tables is not None:
        user_tables = u_tables.split(",")
        user_tables = [table.strip(' ') for table in user_tables]

    # If given from user, then only user tables should be consider
    tables_to_gen = list()
    if user_tables:
        for table_name in db_tables:
            if table_name in user_tables:
                tables_to_gen.append(table_name)
    else:
        tables_to_gen = db_tables

    return tables_to_gen


def check_constraints(tables_to_sort: List[str], db_constraints: Optional[List[Tuple[str]]] = None,
                      cursor_target: cursor = None, connection_target=None):

    # Only constraint relation of given tables should be considered
    constraints: List[Tuple[str]] = list()
    constraints_to_delete: Dict = {}
    referenced_pk_table: Dict = {}
    for constraint in db_constraints:
        append = True
        if constraint[1] in tables_to_sort and constraint[3] in tables_to_sort \
                and constraint[4] == "FOREIGN KEY":
            # Avoid duplicate entries due to multiple relation between tables
            for duplicate in constraints:
                if constraint[1] != duplicate[1] or constraint[3] != duplicate[3]:
                    continue
                append = False
            if append:
                constraints.append(constraint)
        # Get constraint name to delete in target database
        elif constraint[1] in tables_to_sort and constraint[3] not in tables_to_sort \
                and constraint[4] == "FOREIGN KEY":
            constraints_to_delete[constraint[5]] = constraint[1]
            referenced_pk_table[constraint[5]] = constraint[3]

    if bool(constraints_to_delete):
        print(f"Attention!\n"
              f"The tables you want to synthesize contain foreign key dependencies"
              " to other tables that were not specified.\n"
              "Following constraints are affected and would have to be deleted"
              " in order to generate the data into the specified tables:\n")
        for constraint, table in constraints_to_delete.items():
            print(
                f"""  > constraints_name: {constraint}  - corresponding table: {table}""")

        proceed = str(input("\nDo you wish to proceed with deleting these constraints?"
                            " (YES or NO) ")).upper()
        if proceed == 'YES':
            postgres.delete_constraint(
                constraints_to_delete, cursor_target, connection_target)
            print("\n")
        else:
            sys.exit(
                f"\nPlease specify all dependent tables or remove the corresponding constraints.")

    return constraints


def sort_tables_topological(constraints: List[Tuple[str, str, str, str, str, str]], tables_to_sort: List[str]) -> List[str]:
    # If constraints, then generate a graph to get the tables sorted topologically
    if constraints:
        vertices = len(constraints)
        g = Graph(vertices)
        tables_sorted = g.sort_list(constraints, tables_to_sort)
    else:
        tables_sorted = tables_to_sort

    return tables_sorted


class Graph:
    def __init__(self, vertices: int):
        self.graph = defaultdict(list)  # Dictionary containing adjacency list
        self.V = vertices  # No. of vertices

    def sort_list(self, constraint_list: List[str], tables_to_sort: List[str]) -> List[str]:
        for vertices in constraint_list:
            if "FOREIGN KEY" == vertices[4]:
                u = vertices[1]  # FK
                v = vertices[3]  # Related PK
                # For u, v in constraint_list:
                self.addEdge(u, v)
        sorted_list = self.topological_sort(tables_to_sort)
        return sorted_list

    def addEdge(self, u: str, v: str):
        self.graph[u].append(v)

    def topological_sort(self, tables_to_sort: List[str]) -> List[str]:
        sorted_list = list()
        pk_generated = None
        counter = 0
        # Add tables which are not referenced (independent)
        for table in tables_to_sort:
            counter = counter + 1
            if table not in self.graph.keys():
                if table not in sorted_list:
                    sorted_list.insert(0, table)

        # Add tables which are referenced (dependent)
        while len(sorted_list) < len(tables_to_sort):
            pre_counter = len(sorted_list)
            for table in tables_to_sort:
                for child in self.graph.keys():  # Referencing table (child table)
                    if table == child:
                        pk_generated = True
                        # Referenced tables (parent tables)
                        for parent in self.graph[child]:
                            if parent in sorted_list:
                                continue
                            pk_generated = False  # Referenced table not in list yet

                if pk_generated and table not in sorted_list:
                    sorted_list.insert(0, table)
            # Check if sorted list increased
            if pre_counter == len(sorted_list) and len(sorted_list) < len(tables_to_sort):
                sys.exit("""The specified tables are not topologically sortable as they contain interdependencies.
                            Please remove the necessary dependencies between the tables
                            so that the relations (constraints) do not represent a cycle..""")

        return sorted_list[::-1]



from typing import List, Any
from src.generator import numeric_generator, geometry_generator, bytea_generator, date_generator, varchar_generator


def generate_no_constraint(data_type: str, rows_to_gen: int, avg_width: int,
                           table_name: str, numeric_precision: int = None,
                           numeric_precision_radix: int = None, numeric_scale: int = None,
                           min_bound: Any = None, max_bound: Any = None,
                           list_for_bound: List[str] = None, histogram: List[any] = None) -> List[str]:

    generated_vals = list()

    def numeric():
        nonlocal generated_vals
        generated_vals = numeric_generator.get_numeric_vals_no_constraints(data_type, rows_to_gen,
                                                                           min_bound, max_bound,
                                                                           numeric_precision, numeric_precision_radix, numeric_scale,
                                                                           histogram)

    def date():
        nonlocal generated_vals
        generated_vals = date_generator.get_data_no_constraints(
            data_type, rows_to_gen, min_bound, max_bound)

    def varchar():
        nonlocal generated_vals
        generated_vals = varchar_generator.get_varchar_no_constraints(
            rows_to_gen, avg_width, list_for_bound)

    def geometry():
        nonlocal generated_vals
        generated_vals = geometry_generator.get_geometry_no_constraints(
            data_type, rows_to_gen, max_bound)

    def bytea():
        nonlocal generated_vals
        generated_vals = bytea_generator.get_bytea_no_constraints(rows_to_gen)

    def boolean():
        return

    def no_type():
        print(
            f'The "{data_type}" data type is not supported. '
            f'Skipping the table\'s "{table_name}" data generation...')
        return

    funcDict = {'text': varchar,
                'character varying': varchar,
                'character': varchar,
                'smallint': numeric,
                'integer': numeric,
                'bigint': numeric,
                'decimal': numeric,
                'numeric': numeric,
                'double precision': numeric,
                'real': numeric,
                'date': date,
                'timestamp': date,
                'timestamp without time zone': date,
                'point': geometry,
                'line': geometry,
                'polygon': geometry,
                'circle': geometry,
                'box': geometry,
                'lseg': geometry,
                'path': geometry,
                'bytea': bytea,
                'boolean': boolean,
                'not_supported': no_type}

    funcDict[data_type]()

    return generated_vals


def generate_fk(data_type: str, rows_to_gen: int, table_name: str, column_name: str,
                pk_values, fk_ref_info, min_bound: Any = None,
                max_bound: Any = None, histogram: List[any] = None) -> List[Any]:

    def numeric():
        nonlocal generated_vals
        generated_vals = numeric_generator.get_numeric_vals_foreign_constraint(data_type,
            rows_to_gen, pk_values, ref_table, ref_column, min_bound, max_bound, histogram)

    def date():
        nonlocal generated_vals
        generated_vals = date_generator.get_data_foreign_constraint(
            rows_to_gen, pk_values, ref_table, ref_column, min_bound, max_bound)

    def varchar():
        nonlocal generated_vals
        generated_vals = varchar_generator.get_varchar_foreign_constraint(
            rows_to_gen, pk_values, ref_table, ref_column)

    def no_type():
        print(
            f'The "{data_type}" data type is not supported. '
            f'Skipping the table\'s "{table_name}" data generation...')
        return

    generated_vals = list()

    ref_table = fk_ref_info[column_name]['ref_table']
    ref_column = fk_ref_info[column_name]['ref_column']

    funcDict = {'text': varchar,
                'character varying': varchar,
                'character': varchar,
                'smallint': numeric,
                'integer': numeric,
                'bigint': numeric,
                'decimal': numeric,
                'numeric': numeric,
                'date': date,
                'timestamp': date,
                'timestamp without time zone': date,
                'not_supported': no_type}

    funcDict[data_type]()

    return generated_vals


def generate_pk(data_type: str, rows_to_gen: int, avg_width: int, table_name: str,
                min_value: Any = None, max_value: Any = None,
                list_for_bound: List[str] = None, histogram: List[any] = None
                ) -> List[Any]:

    generated_vals = list()

    def numeric():
        nonlocal generated_vals
        generated_vals = numeric_generator.get_numeric_vals_primary_constraint(data_type,
            rows_to_gen, min_value, max_value, histogram)

    def date():
        nonlocal generated_vals
        generated_vals = date_generator.get_date_primary_constraints(
            data_type, rows_to_gen, min_value, max_value)

    def varchar():
        nonlocal generated_vals
        generated_vals = varchar_generator.get_varchar_primary_constraints(
            rows_to_gen, avg_width, list_for_bound)

    def no_type():
        """ print(
            f'The "{data_type}" data type is not supported. '
            f'Skipping the table\'s "{table_name}" data generation...') """
        return

    funcDict = {'text': varchar,
                'character varying': varchar,
                'character': varchar,
                'smallint': numeric,
                'integer': numeric,
                'bigint': numeric,
                'decimal': numeric,
                'numeric': numeric,
                'date': date,
                'timestamp': date,
                'timestamp without time zone': date,
                'not_supported': no_type}

    funcDict[data_type]()

    return generated_vals



from src import utils
from src import model_generator
import datetime


def test_number_bound_mvc():
    # Testing requirement MM-3
    cursor_src = None
    table_name = 'atp_players'
    column_name = 'player_id'
    data_type = 'integer'
    most_common_values = [2, 3, 4, 5, 6, 6, 7, 5, 34, 2, 34, 2, 33, 4]
    histogram_bounds = [11, 22, 33, 44, 55, 66, 77, 88, 99, 12, 23]

    min_bound, max_bound, list_for_bound = model_generator.determine_boundaries(cursor_src, table_name,
                                                                                column_name, data_type,
                                                                                most_common_values,
                                                                                histogram_bounds)
    assert min_bound == 2
    assert max_bound == 34
    assert list_for_bound == [2, 3, 4, 5, 6, 6, 7, 5, 34, 2, 34, 2, 33, 4]


def test_number_bound_hist():
    # MM-4
    cursor_src = None
    table_name = 'atp_players'
    column_name = 'player_id'
    data_type = 'integer'
    most_common_values = None
    histogram_bounds = [11, 22, 33, 44, 55, 66, 77, 88, 99, 12, 23]

    min_bound, max_bound, list_for_bound = model_generator.determine_boundaries(cursor_src, table_name,
                                                                                column_name, data_type,
                                                                                most_common_values,
                                                                                histogram_bounds)
    assert min_bound == 11
    assert max_bound == 99
    assert list_for_bound == [11, 22, 33, 44, 55, 66, 77, 88, 99, 12, 23]


def test_date_bound_mcv():

    cursor_src = None
    table_name = 'atp_players'
    column_name = 'player_id'
    data_type = 'date'
    most_common_values = ['2025-04-01',
                          '2010-03-02', '2011-1-03', '2020-03-18']
    histogram_bounds = ['2034-01-12', '1919-04-22',
                        '2028-12-01', '2030-08-08', '1999-01-01']

    min_bound, max_bound, list_for_bound = model_generator.determine_boundaries(cursor_src, table_name,
                                                                                column_name, data_type,
                                                                                most_common_values,
                                                                                histogram_bounds)
    assert min_bound == datetime.date(year=2010, month=3, day=2)
    assert max_bound == datetime.date(year=2025, month=4, day=1)
    assert list_for_bound == ['2025-04-01',
                              '2010-03-02', '2011-1-03', '2020-03-18']


def test_date_bound_hist():

    cursor_src = None
    table_name = 'atp_players'
    column_name = 'player_id'
    data_type = 'date'
    most_common_values = None
    histogram_bounds = ['2025-04-01', '2010-03-02', '2011-1-03', '2020-03-18']

    min_bound, max_bound, list_for_bound = model_generator.determine_boundaries(cursor_src, table_name,
                                                                                column_name, data_type,
                                                                                most_common_values,
                                                                                histogram_bounds)
    assert min_bound == datetime.date(year=2010, month=3, day=2)
    assert max_bound == datetime.date(year=2025, month=4, day=1)
    assert list_for_bound == ['2025-04-01',
                              '2010-03-02', '2011-1-03', '2020-03-18']


def test_date_bound_no_list():
    # MM-6
    cursor_src = None
    table_name = 'atp_players'
    column_name = 'player_id'
    data_type = 'date'
    most_common_values = None
    histogram_bounds = None

    min_bound, max_bound, list_for_bound = model_generator.determine_boundaries(cursor_src, table_name,
                                                                                column_name, data_type,
                                                                                most_common_values,
                                                                                histogram_bounds)
    assert min_bound == datetime.date(year=1950, month=1, day=1)
    assert max_bound == datetime.date.today()
    assert list_for_bound is None


def test_get_date_boundaries():
    # helper function in utils
    mcv = ['2020-01-01', '2020-01-02', '2020-01-03', '2020-01-04']
    min = utils.get_date_bound(mcv, 'start')
    max = utils.get_date_bound(mcv, 'end')
    assert min == datetime.date(year=2020, month=1, day=1)
    assert max == datetime.date(year=2020, month=1, day=4)



import numpy as np
from typing import Any


def get_geometry_no_constraints(data_type: str, rows_to_gen: int, max_bound: Any = None):
    generated_vals = list()

    if max_bound is None:
        max_bound = 1000

    if data_type == 'point':
        for _ in range(rows_to_gen):
            generated_vals.append(
                f'point({np.random.randint(1,max_bound)},{np.random.randint(1,max_bound)})')

    if data_type == 'line':
        for _ in range(rows_to_gen):
            value = f'line( point \'({np.random.randint(1,max_bound)}, \
            {np.random.randint(1,max_bound)})\',\'({np.random.randint(1,max_bound)}, \
            {np.random.randint(1,max_bound)})\')'
            generated_vals.append(value)

    if data_type == 'polygon':
        for _ in range(rows_to_gen):
            start_0 = np.random.randint(1, max_bound)
            start_1 = np.random.randint(1, max_bound)
            generated_vals.append(
                f' polygon \'({start_0,start_1},({np.random.randint(1,max_bound)}, \
                {np.random.randint(1,max_bound)}),{start_0,start_1})\'')

    if data_type == 'circle':
        for _ in range(rows_to_gen):
            generated_vals.append(
                f'circle \'<({np.random.randint(1,max_bound)}, \
                {np.random.randint(1,max_bound)}),{np.random.randint(1,10)}>\'')

    if data_type == 'box':
        for _ in range(rows_to_gen):
            generated_vals.append(
                'box \'(({0},{1}),({2},{3}))\''.format(
                    np.random.randint(1, max_bound),
                    np.random.randint(1, max_bound),
                    np.random.randint(1, max_bound),
                    np.random.randint(1, max_bound)))

    if data_type == 'lseg':
        for _ in range(rows_to_gen):
            generated_vals.append(
                'lseg \'(({0},{1}),({2},{3}))\''.format(
                    np.random.randint(1, max_bound),
                    np.random.randint(1, max_bound),
                    np.random.randint(1, max_bound),
                    np.random.randint(1, max_bound)))

    if data_type == 'path':
        for _ in range(rows_to_gen):
            start_0 = np.random.randint(1, max_bound)
            start_1 = np.random.randint(1, max_bound)
            generated_vals.append(
                f' path \'({start_0,start_1},({np.random.randint(1,max_bound)},\
                {np.random.randint(1,max_bound)}),  \
                {start_0,start_1})\'')

    return generated_vals



import datetime
import numpy
from src import data_generator
from src import postgres
from src import utils
from src import query_generator
from psycopg2._psycopg import cursor
from typing import List, Dict, Tuple, Any, Optional


DEFAULT_NUMBER_OF_ROWS = 100
RANDOM_WORD_LENGTH = 15

# TODO: eliminate global variables
table_information: Dict = {}
pk_values: Dict = {}
fk_ref_info: Dict = {}


def generate_model(cursor_src, table_name, tables_sorted, db_constraints, mf=None) -> Dict:

    # TODO: Create more functions!
    print(f'Generating synthetic data for the "{table_name}" table...')
    insert_query = ""

    # Determine number of rows in source table
    cursor_src.execute(f"SELECT COUNT(*) FROM {table_name};")
    number_of_rows = cursor_src.fetchone()[0]
    if number_of_rows is None:
        number_of_rows = DEFAULT_NUMBER_OF_ROWS

    # Consider multiplication factor
    total_rows = (int(number_of_rows * mf))

    # Fill dict with general and statistical table information
    collect_meta_data(cursor_src, table_name)

    # Get tables column names
    column_names = list()

    for column_name, column_info in table_information.get(table_name)["column_information"].items():
        """ The next line messes up the primary key insertion. It forces to diregard 
                the generated_vals and does an auto increment where it is specified as a 
                column_default. 

                This can subsequently lead to insertion errors on foreign keys since primary 
                keys might not be the same in generated database and the ones used to 
                generate foreign key values in the tool. 

            """
        # if not column_info.get("column_default"):
        column_names.append(column_name)

    if not column_names:
        print(f'No columns found to generate data into. ' f'Skipping the'
              f' table\'s "{table_name}" data generation...')

    # Retrieve column general information
    column_information: Dict = {}
    for column_info in table_information[table_name]["column_information"].values():
        column_name = column_info.get("column_name")
        data_type = column_info.get("data_type")
        numeric_precision = column_info.get("numeric_precision")
        numeric_precision_radix = column_info.get("numeric_precision_radix")
        numeric_scale = column_info.get("numeric_scale")
        column_information[column_name] = {}

        if data_type not in postgres.DataTypes.SUPPORTED_TYPES:
            data_type = 'not_supported'

        # Retrieve column statistical information
        if column_name in table_information[table_name]["pg_stats"]:
            column_stats: Dict = table_information[table_name]["pg_stats"][column_name]
            most_common_values = column_stats["most_common_vals"]
            most_common_freqs = column_stats["most_common_freqs"]
            avg_width = column_stats["avg_width"]
            n_distinct = column_stats["n_distinct"]
            histogram_bounds = column_stats["histogram_bounds"]

            # Determine_frequency
            generated_freqs, rows_to_gen = determine_frequency(number_of_rows,
                                                               total_rows,
                                                               most_common_values,
                                                               most_common_freqs,
                                                               n_distinct)

            # Determine_boundaries
            min_bound, max_bound, list_for_bound = determine_boundaries(cursor_src, table_name,
                                                                        column_name, data_type,
                                                                        most_common_values,
                                                                        histogram_bounds)

            # Determine whether column is constraint
            constraint = determine_constraint_info(
                table_name, column_name, tables_sorted, db_constraints)
            if constraint == "FOREIGN KEY":
                generated_vals = data_generator.generate_fk(data_type, rows_to_gen, table_name,
                                                            column_name, pk_values,
                                                            fk_ref_info, min_bound, max_bound, histogram_bounds)

            elif constraint == "PRIMARY KEY":
                generated_vals = data_generator.generate_pk(data_type, rows_to_gen, avg_width, table_name,
                                                            min_bound, max_bound, list_for_bound, histogram_bounds)

                pk_values[table_name][column_name] = generated_vals

            else:
                generated_vals = data_generator.generate_no_constraint(data_type, rows_to_gen, avg_width,
                                                                       table_name, numeric_precision,
                                                                       numeric_precision_radix, numeric_scale,
                                                                       min_bound, max_bound, list_for_bound, histogram_bounds)

            column_information[column_name]["generated_vals"] = generated_vals
            column_information[column_name]["generated_freqs"] = generated_freqs

    insert_dict: Dict = {}
    insert_dict = query_generator.create_insert_query(total_rows, insert_query, table_name, column_names,
                                                      table_information, column_information,
                                                      insert_dict)

    return insert_dict


def collect_meta_data(cursor_src, table_name: str) -> None:

    # Initialize dict per table
    table_information[table_name] = {}
    table_information[table_name]["column_information"] = {}
    table_information[table_name]["pg_stats"] = {}

    # Retrieve general column information and fill into dict
    fill_columns_dict(cursor_src, table_name)

    # Retrieve statistical column information and fill into dict
    fill_stats_dict(cursor_src, table_name)


def fill_columns_dict(cursor_src, table_name: str) -> None:

    column_results = postgres.get_column_information(cursor_src, table_name)

    for column_entry in column_results:
        columns_dict = dict()
        columns_dict["column_name"] = column_entry[0]
        columns_dict["data_type"] = column_entry[1]
        columns_dict["max_length"] = column_entry[2]
        if column_entry[3]:
            columns_dict["column_default"] = column_entry[3]
        if column_entry[4]:
            columns_dict["numeric_precision"] = column_entry[4]
        if column_entry[5]:
            columns_dict["numeric_precision_radix"] = column_entry[5]
        if column_entry[6]:
            columns_dict["numeric_scale"] = column_entry[6]

        table_information[table_name]["column_information"][column_entry[0]] = columns_dict


def fill_stats_dict(cursor_src, table_name: str) -> None:

    table_stats = postgres.get_table_stats(cursor_src, table_name)
    for stats_entry in table_stats:
        stats_dict = dict()
        stats_dict["column_name"] = stats_entry[0]
        stats_dict["null_frac"] = stats_entry[1]
        stats_dict["avg_width"] = stats_entry[2]
        stats_dict["n_distinct"] = stats_entry[3]

        most_common_values = stats_entry[4]
        if most_common_values is not None:
            most_common_values = most_common_values.strip("{}").split(",")
            most_common_values = [value.strip('"').replace(
                "'", "''") for value in most_common_values]
            most_common_values = [
                value for value in most_common_values if value.strip()]

        stats_dict["most_common_vals"] = most_common_values
        stats_dict["most_common_freqs"] = stats_entry[5]

        histogram_bounds = stats_entry[6]
        if histogram_bounds is not None:
            histogram_bounds = histogram_bounds.strip("{}").split(",")
            histogram_bounds = [bound.strip('"').replace(
                "'", "''") for bound in histogram_bounds]
            histogram_bounds = [
                bound for bound in histogram_bounds if bound.strip()]

        stats_dict["histogram_bounds"] = histogram_bounds
        stats_dict["correlation"] = stats_entry[7]
        table_information[table_name]["pg_stats"][stats_entry[0]] = stats_dict


def determine_frequency(number_of_rows: int, total_rows: int, most_common_values: List[str] = None,
                        most_common_freqs: List[float] = None, n_distinct: float = None) -> Tuple[float, int]:

    if most_common_values and most_common_freqs:

        if n_distinct > 0:
            distinct_no = n_distinct
        else:
            distinct_no = -n_distinct * number_of_rows
        distinct_no = round(distinct_no)
        leftover_freq = 1 - sum(most_common_freqs)
        generated_freqs = most_common_freqs

        # The dirichlet function that generates random floating numbers to fill
        # The left-over frequencies
        # Generated_freqs indicates the probability for a value to get picked
        generated_freqs += (numpy.random.dirichlet(numpy.ones(distinct_no - len(most_common_freqs)))
                            * leftover_freq).tolist()

        # Rows_to_gen has to be equal as distinct values
        rows_to_gen = len(generated_freqs)

    # No mcv means that no values seem to be more common than any others (all unique)
    else:
        # Rows_to_gen has to be equal as rows_to_gen since all unique
        generated_freqs = False
        rows_to_gen = total_rows

    return generated_freqs, rows_to_gen


def determine_boundaries(cursor_src: cursor, table_name: str, column_name: str, data_type: str,
                         most_common_values: Optional[List[any]] = None,
                         histogram_bounds: Optional[List[any]] = None) -> Tuple[Any, Any, Optional[List[any]]]:

    list_for_bound: Optional[List[any]
                             ] = most_common_values or histogram_bounds or None

    def numeric():
        if list_for_bound:
            if data_type in ('decimal', 'numeric'):
                min_value = float(min(list_for_bound))
                max_value = float(max(list_for_bound))
            else:
                list_for_bound_int = list()
                for max_str in list_for_bound:
                    max_int = int(max_str)
                    list_for_bound_int.append(max_int)
                min_value = min(list_for_bound_int)
                max_value = max(list_for_bound_int)

        else:
            min_value = None
            cursor_src.execute(f"SELECT MIN({column_name}) FROM {table_name}")
            result = cursor_src.fetchone()
            if result:
                min_value = result[0]
            max_value = None
            cursor_src.execute(f"SELECT MAX({column_name}) FROM {table_name}")
            result = cursor_src.fetchone()
            if result:
                max_value = result[0]

        return min_value, max_value

    def date():
        if list_for_bound:
            START_DATE = utils.get_date_bound(list_for_bound, "start")
            END_DATE = utils.get_date_bound(list_for_bound, "end")
        else:
            START_DATE = datetime.date(year=1950, month=1, day=1)
            END_DATE = datetime.date.today()

        return START_DATE, END_DATE

    def geometry():
        min_value = max_value = 1000
        return min_value, max_value

    def bytea():
        min_value = max_value = 8
        return min_value, max_value

    def no_type():
        # Data type not supported or boundaries not necessary
        min_value = max_value = None
        return min_value, max_value

    funcDict = {
        'smallint': numeric,
        'integer': numeric,
        'bigint': numeric,
        'decimal': numeric,
        'numeric': numeric,
        'date': date,
        'timestamp': date,
        'timestamp without time zone': date,
        'point': geometry,
        'line': geometry,
        'polygon': geometry,
        'circle': geometry,
        'box': geometry,
        'lseg': geometry,
        'path': geometry,
        'bytea': bytea,
        'not_supported': no_type}

    if data_type in funcDict.keys():
        min_bound, max_bound = funcDict[data_type]()
    else:
        min_bound, max_bound = no_type()

    return min_bound, max_bound, list_for_bound


def determine_constraint_info(table_name: str, column_name: str, tables_sorted: List[str],
                              db_constraints: List[Tuple[str]]) -> str:

    # Retrieve column constraint information
    for key in db_constraints:
        if column_name == key[0] and table_name == key[1] and "FOREIGN KEY" == key[4] \
                and key[3] in tables_sorted:
            constraint = "FOREIGN KEY"
            fk_ref_info[column_name] = dict()
            fk_ref_info[column_name]['ref_column'] = key[2]
            fk_ref_info[column_name]['ref_table'] = key[3]

            return constraint

        elif column_name == key[0] and table_name == key[1] and "PRIMARY KEY" == key[4]:
            constraint = "PRIMARY KEY"
            pk_values[table_name] = {}
            pk_values[table_name][column_name] = dict()

            return constraint



import os
import subprocess
import sys
from subprocess import Popen
from typing import List, Optional, Any
import psycopg2
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
from psycopg2._psycopg import cursor
from psycopg2 import sql
from src import postgres
from src import model_generator
from src import utils
import argparse
from argparse import Namespace

__version__ = '2.0'
examples: Optional[str] = '''How to use pgsynthdata.py:

  python pgsynthdata.py test postgres -show
  \t-> Connects to database "test", host="localhost", port="5432", default user with password "postgres"
  \t-> Shows statistics from the tables in database "test"
  
  python pgsynthdata.py db pw1234 -H myHost -p 8070 -U testuser -show
  \t-> Connects to database "db", host="myHost", port="8070", user="testuser" with password "pw1234"
  \t-> Shows statistics from the tables in database "db"
  
  python pgsynthdata.py dbin dbgen pw1234 -H myHost -p 8070 -U testuser -generate
  \t-> Connects to database "dbin", host="myHost", port="8070", user="testuser" with password "pw1234"
  \t-> Creates new database "dbgen" with synthetic data
  
  python pgsynthdata.py dbin dbgen pw1234 -H myHost -p 8070 -U testuser -generate -tables table1,table2
  \t-> Connects to database "dbin", host="myHost", port="8070", user="testuser" with password "pw1234"
  \t-> Creates new database "dbgen" with synthetic data on tables: "table1" and "table2"
  
  python pgsynthdata.py --version
  \t-> Show the version of this program and quit'''

DUMP_FILE_PATH = 'schema.dump'


def main() -> None:
    args = parse_arguments()

    if args.show:
        show(args)
    else:
        if args.DB_NAME_GEN is None:
            sys.exit(
                'When "-generate" argument is given, the following argument is required: DB_NAME_GEN')
        else:
            connection = None
            try:

                connection = psycopg2.connect(dbname=args.DB_NAME_IN,
                                              user=args.user,
                                              host=args.hostname,
                                              port=args.port,
                                              password=args.password)

                connection.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
                cursor = connection.cursor()
                postgres.analyze_database(cursor, args.DB_NAME_IN)

                generate(connection, cursor, args)
            except psycopg2.DatabaseError:
                sys.exit('''Connection failed because of at least one of the following reasons:
                        Database does not exist
                        User does not exist
                        Wrong password''')
            finally:
                if connection is not None:
                    connection.close()


def parse_arguments() -> Namespace:
    parser = argparse.ArgumentParser(description='Connects to a database and reads statistics',
                                     epilog=examples,
                                     formatter_class=argparse.RawDescriptionHelpFormatter)

    parser.add_argument('-v', '--version', action='version',
                        version=f'pgsynthdata version: {__version__}')
    parser.add_argument('DB_NAME_IN', type=str,
                        help='Name of an existing postgres database')
    parser.add_argument('DB_NAME_GEN', type=str, nargs='?',
                        help='Name of database to be created')  # optional, but not if DB_NAME_GEN is given
    parser.add_argument('password', type=str, help='Required user password')

    # One of the two options in action_group has to be given, but not both
    action_group = parser.add_mutually_exclusive_group(required=True)
    action_group.add_argument(
        '-show', '--show', action='store_true', help='If given, shows config')
    action_group.add_argument('-generate', '--generate', action='store_true',
                              help='If given, generates new synthesized data to database DB_NAMEG_EN')

    parser.add_argument('-mf', '--mf', type=float, default=1.0,
                        help='Multiplication factor (mf) for the generated synthesized data (default: 1.0)')
    parser.add_argument('-tables', '--tables', type=str,
                        help='Only generate data for specific tables, separated by a comma')

    parser.add_argument('-O', '--owner', type=str,
                        help='Owner of the database, default: same as user')
    parser.add_argument('-H', '--hostname', type=str,
                        help='Specifies the host name, default: localhost')
    parser.add_argument('-p', '--port', type=int,
                        help='Specifies the TCP/IP port, default: 5432')
    parser.add_argument('-U', '--user', type=str,
                        help='An existing postgres database user, default: default user')

    return parser.parse_args()


def show(args: Any) -> None:
    connection = None
    try:
        connection = psycopg2.connect(dbname=args.DB_NAME_IN,
                                      user=args.user,
                                      host=args.hostname,
                                      port=args.port,
                                      password=args.password)

        cursor = connection.cursor()

        postgres.show_database_stats(cursor, args.tables)
        cursor.close()

    except psycopg2.DatabaseError:
        sys.exit('''Connection failed because of at least one of the following reasons:
                    Database does not exist
                    User does not exist
                    Wrong password''')
    finally:
        if connection is not None:
            connection.close()


def generate(connection, cursor: cursor, args) -> None:

    print(
        f'Preparing the generation of synthetic data into the "{args.DB_NAME_GEN}" database...\n')

    # Build up new Database based on source database DB_NAME_IN
    postgres.create_database(connection, cursor, args.DB_NAME_GEN, args.owner)
    copy_database_structure(args)
    cursor.close()

    # Connect to source and target database
    connection_src, cursor_src = postgres.get_db_cursor(
        args.DB_NAME_IN, args.user, args.hostname, args.port, args.password)
    connection_target, cursor_target = postgres.get_db_cursor(
        args.DB_NAME_GEN, args.user, args.hostname, args.port, args.password)

    # Collect all constraints
    db_constraints = postgres.get_constr_information(cursor_src)

    # Collect all table names from source database
    table_results = postgres.get_tables(cursor_src)

    # Arrange tables based on constraints
    tables_to_sort = utils.determine_tables_to_generate(
        table_results, args.tables)

    constraints_to_consider = utils.check_constraints(tables_to_sort, db_constraints,
                                                      cursor_target, connection_target)

    tables_sorted = utils.sort_tables_topological(
        constraints_to_consider, tables_to_sort)

    insert_column_settings(cursor_target, connection_target, tables_sorted)

    for table_name in tables_sorted:
        # Generate data model and data based on input data
        insert_dict = model_generator.generate_model(cursor_src=cursor_src,
                                                     table_name=table_name,
                                                     tables_sorted=tables_sorted,
                                                     db_constraints=db_constraints,
                                                     mf=args.mf)

        # Insert table data into targe database DB_NAME_GEN
        insert_data(cursor_target, connection_target, insert_dict)

    cursor_src.close()
    cursor_target.close()
    sys.stdout.write(
        f'Successfully generated the synthetic data into the "{args.DB_NAME_GEN}" database.')


def copy_database_structure(args) -> None:
    print(f'Copying the "{args.DB_NAME_GEN}" database structure...\n')

    try:
        process = Popen(['pg_dump',
                         '--dbname=postgresql://{}:{}@{}:{}/{}'.format(args.user,
                                                                       args.password,
                                                                       'localhost',
                                                                       '5432',
                                                                       args.DB_NAME_IN),
                         '-s',
                         '-Fc',
                         '-f', DUMP_FILE_PATH
                         ],
                        stdout=subprocess.PIPE)

        process.communicate()[0]

        process = Popen(['pg_restore',
                         '--dbname=postgresql://{}:{}@{}:{}/{}'.format(args.user,
                                                                       args.password,
                                                                       'localhost',
                                                                       '5432',
                                                                       args.DB_NAME_GEN),
                         DUMP_FILE_PATH],
                        stdout=subprocess.PIPE
                        )

        process.communicate()[0]
    except Exception as error:
        sys.exit('Database structure could not be copied. Error: {}'.format(error))
    finally:
        if os.path.exists(DUMP_FILE_PATH):
            os.remove(DUMP_FILE_PATH)


def insert_data(cursor_target: cursor, connection_target, insert_dict) -> None:
    for table_name, insert_query in insert_dict.items():
        print(f'Inserting synthetic data into the "{table_name}" table...\n')
        try:
            if insert_query:
                cursor_target.execute(sql.SQL(insert_query).format(
                    table_name=sql.Identifier(table_name)))
                connection_target.commit()

        except psycopg2.DatabaseError as db_error:
            sys.stdout.write(
                f'An error occurred while inserting data into the "{table_name}" table.'
                f' Error description: {db_error}.\n')
            connection_target.rollback()


def insert_column_settings(cursor_target: cursor, connection_target, tables: List[any]) -> None:
    print(f'Gathering and inserting information_schema for sorted tables')
    string_list = '('
    counter = 0
    for table in tables:
        counter += 1
        string_list += '\''
        string_list += table
        string_list += '\''
        if counter != len(tables):
            string_list += ','
    string_list += ')'

    tab_var_setting = 'tab_var_setting'
    insert_query = f'select * into {tab_var_setting} from information_schema.columns where true and table_name in {string_list} order by ordinal_position'

    try:
        cursor_target.execute(sql.SQL(insert_query).format(
            table_name=sql.Identifier(tab_var_setting)))
        connection_target.commit()
    except psycopg2.DatabaseError as db_error:
        sys.stdout.write(
            f'An error occurred while inserting data into the "{tab_var_setting}" table.'
            f' Error description: {db_error}.\n')
        connection_target.rollback()


if __name__ == '__main__':
    main()



from src.generator import generator_utils
from typing import Any
import datetime


def get_data_no_constraints(data_type: str, rows_to_gen: int, min_bound: Any = None, max_bound: Any = None):

    generated_vals = list()
    for index in range(rows_to_gen):
        if data_type in ('timestamp', 'timestamp without time zone'):
            generated_vals.append(generator_utils.random_date(
                min_bound, max_bound, time=True))
        else:
            generated_vals.append(
                generator_utils.random_date(min_bound, max_bound))

    return generated_vals


def get_data_foreign_constraint(rows_to_gen: int,
                                pk_values, ref_table, ref_column, min_bound: Any = None,
                                max_bound: Any = None):

    generated_vals = list()
    counter = 0
    pk_values_list = pk_values[ref_table][ref_column]
    while len(generated_vals) < rows_to_gen:  # ensure that all values are created
        for value in pk_values_list:
            # ensure not to exceed values during for loop
            if len(generated_vals) < rows_to_gen:
                # consider every value in PK_values once
                if counter < len(pk_values_list):
                    counter = counter + 1
                    if min_bound <= value <= max_bound and value not in generated_vals:
                        generated_vals.append(value)
                else:  # if pk_values does not contain enough values then extend boundaries
                    l_min_bound = min(pk_values_list)
                    l_max_bound = max(pk_values_list)
                    if l_min_bound <= value <= l_max_bound and value not in generated_vals:
                        generated_vals.append(value)

    return generated_vals


def get_date_primary_constraints(data_type, rows_to_gen: int, min_value: Any = None, max_value: Any = None):
    generated_vals = list()
    # ensure that enough dates btw. start and end are available
    days_between = generator_utils.days_between(min_value, max_value)
    if rows_to_gen > days_between:
        max_value = min_value + datetime.timedelta(days=rows_to_gen)
    while len(generated_vals) < rows_to_gen:
        for index in range(rows_to_gen):
            if data_type in ('timestamp', 'timestamp without time zone'):
                value = generator_utils.random_date(
                    min_value, max_value, time=True)
                if value not in generated_vals:
                    generated_vals.append(value)
            else:
                value = generator_utils.random_date(min_value, max_value)
                if value not in generated_vals:
                    generated_vals.append(value)

    return generated_vals



import pytest
from src import utils
from src import model_generator


def test_topological_sorting():
    # Sort tables topological
    tables_to_sort = ['atp_players', 'atp_matches', 'atp_rankings']
    constraints_to_consider = [('loser_id', 'atp_matches', 'player_id', 'atp_players', 'FOREIGN KEY', 'atp_matches_loser_id_fkey'),
                               ('player_id', 'atp_rankings', 'player_id', 'atp_players',
                                'FOREIGN KEY', 'atp_rankings_player_id_fkey'),
                               ('ranking_date', 'atp_jury', 'ranking_date', 'atp_rankings', 'FOREIGN KEY', 'atp_jury_ranking_date_fkey')]

    tables_sorted = utils.sort_tables_topological(
        constraints_to_consider, tables_to_sort)
    print(tables_sorted)
    assert tables_sorted == ['atp_players', 'atp_matches', 'atp_rankings']


def test_topological_sorting_cycle_negative():
    # Given tables are not topologically sortable, as they represent a cycle, i.e. interdependence
    # Atp_players is dependent on atp_rankings and vice versa
    with pytest.raises(SystemExit) as pytest_wrapped_e:
        tables_to_sort = ['atp_players', 'atp_matches', 'atp_rankings']
        constraints_to_consider = [
            ('loser_id', 'atp_matches', 'player_id', 'atp_players',
             'FOREIGN KEY', 'atp_matches_loser_id_fkey'),
            ('player_id', 'atp_rankings', 'player_id', 'atp_players',
             'FOREIGN KEY', 'atp_rankings_player_id_fkey'),
            ('ranking_date', 'atp_jury', 'ranking_date', 'atp_rankings',
             'FOREIGN KEY', 'atp_jury_ranking_date_fkey'),
            ('player_name', 'atp_players', 'player_id', 'atp_rankings', 'FOREIGN KEY', 'atp_test_key')]

        utils.sort_tables_topological(constraints_to_consider, tables_to_sort)

    assert pytest_wrapped_e.type == SystemExit


def test_constraint_collector():

    tables_to_sort = ['atp_players', 'atp_matches', 'atp_rankings', 'atp_jury']

    db_constraints = [('player_id', 'atp_players', 'player_id', 'atp_players', 'PRIMARY KEY', 'atp_players_pkey'),
                      ('loser_id', 'atp_matches', 'player_id', 'atp_players',
                       'FOREIGN KEY', 'atp_matches_loser_id_fkey'),
                      ('winner_id', 'atp_matches', 'player_id', 'atp_players',
                       'FOREIGN KEY', 'atp_matches_winner_id_fkey'),
                      ('player_id', 'atp_rankings', 'player_id', 'atp_players',
                       'FOREIGN KEY', 'atp_rankings_player_id_fkey'),
                      ('ranking_date', 'atp_rankings', 'ranking_date',
                       'atp_rankings', 'PRIMARY KEY', 'atp_rankings_pkey'),
                      ('ranking_date', 'atp_jury', 'ranking_date', 'atp_rankings',
                       'FOREIGN KEY', 'atp_jury_ranking_date_fkey'),
                      ('jury_name', 'atp_jury', 'jury_name', 'atp_jury', "PRIMARY KEY"    "atp_jury_pkey")]

    constraints_to_consider = utils.check_constraints(
        tables_to_sort, db_constraints, None, None)

    assert constraints_to_consider == [('loser_id', 'atp_matches', 'player_id', 'atp_players', 'FOREIGN KEY', 'atp_matches_loser_id_fkey'),
                                       ('player_id', 'atp_rankings', 'player_id', 'atp_players',
                                        'FOREIGN KEY', 'atp_rankings_player_id_fkey'),
                                       ('ranking_date', 'atp_jury', 'ranking_date', 'atp_rankings', 'FOREIGN KEY', 'atp_jury_ranking_date_fkey')]


def test_determine_constraint_fk():

    table_name = 'atp_matches'
    column_name = 'loser_id'
    tables_sorted = ['atp_players', 'atp_matches']
    db_constraints = [('player_id', 'atp_players', 'player_id', 'atp_players', 'PRIMARY KEY', 'atp_players_pkey'),
                      ('loser_id', 'atp_matches', 'player_id', 'atp_players', 'FOREIGN KEY',
                       'atp_matches_loser_id_fkey'),
                      ('winner_id', 'atp_matches', 'player_id', 'atp_players', 'FOREIGN KEY',
                       'atp_matches_winner_id_fkey')]
    constraint = model_generator.determine_constraint_info(
        table_name, column_name, tables_sorted, db_constraints)
    assert constraint == 'FOREIGN KEY'


def test_determine_constraint_pk():
    table_name = 'atp_players'
    column_name = 'player_id'
    tables_sorted = ['atp_players', 'atp_matches']
    db_constraints = [('player_id', 'atp_players', 'player_id', 'atp_players', 'PRIMARY KEY', 'atp_players_pkey'), ('loser_id', 'atp_matches', 'player_id', 'atp_players',
                                                                                                                    'FOREIGN KEY', 'atp_matches_loser_id_fkey'), ('winner_id', 'atp_matches', 'player_id', 'atp_players', 'FOREIGN KEY', 'atp_matches_winner_id_fkey')]
    constraint = model_generator.determine_constraint_info(
        table_name, column_name, tables_sorted, db_constraints)
    assert constraint == 'PRIMARY KEY'


def test_determine_no_constraint():
    table_name = 'atp_rankings'
    column_name = 'ranking'
    tables_sorted = ['atp_players', 'atp_matches']
    db_constraints = [('player_id', 'atp_players', 'player_id', 'atp_players', 'PRIMARY KEY', 'atp_players_pkey'),
                      ('loser_id', 'atp_matches', 'player_id', 'atp_players',
                       'FOREIGN KEY', 'atp_matches_loser_id_fkey'),
                      ('winner_id', 'atp_matches', 'player_id', 'atp_players', 'FOREIGN KEY', 'atp_matches_winner_id_fkey')]
    constraint = model_generator.determine_constraint_info(
        table_name, column_name, tables_sorted, db_constraints)
    assert constraint is None


def test_determine_tables_to_generate():

    tables_result = [('public', 'atp_jury', '10', '4'),
                     ('public', 'atp_matches', '2781', '2'),
                     ('public', 'atp_players', '54938', '1'),
                     ('public', 'atp_rankings', '47', '3'),
                     ('public', 'atp_sponsor', '3', '5')]

    u_tables = 'atp_matches,atp_rankings,atp_players'

    tables_to_sort = utils.determine_tables_to_generate(
        tables_result, u_tables)

    assert tables_to_sort == ['atp_matches', 'atp_players', 'atp_rankings']



import sys
import psycopg2
from psycopg2 import sql
from psycopg2._psycopg import cursor


class DataTypes:
    VARCHAR_TYPES = [
        'text',
        'character varying',
        'character'
    ]

    NUMERIC_TYPES = [
        'smallint',
        'integer',
        'bigint',
        'decimal',
        'numeric',
        'real',
        'double precision'
    ]

    DATE_TYPES = [
        'date',
        'timestamp',
        'timestamp without time zone'
    ]

    BOOLEAN_TYPES = [
        'boolean',
        'bool'
    ]

    GEOMETRY_TYPES = [
        'point',
        'circle',
        'lseg',
        'polygon',
        'box',
        'line',
        'path'
    ]

    BINARY_TYPES = [
        'bytea'
    ]

    SUPPORTED_TYPES = VARCHAR_TYPES + NUMERIC_TYPES + \
        DATE_TYPES + BOOLEAN_TYPES + GEOMETRY_TYPES + BINARY_TYPES


def get_db_cursor(DB_NAME: str, user: str, hostname: str, port: str, password: str):

    try:
        connection = psycopg2.connect(dbname=DB_NAME,
                                      user=user,
                                      host=hostname,
                                      port=port,
                                      password=password)
    except psycopg2.DatabaseError as error:
        sys.exit('Could not connect to the "{0}" database. Error description: {1}'.format(
            DB_NAME, error))

    cursor = connection.cursor()
    return connection, cursor


def delete_constraint(constraints_to_delete, cursor_target: cursor, connection_target):

    for constraint_name, table_name in constraints_to_delete.items():
        query = sql.SQL("ALTER TABLE {table} DROP CONSTRAINT {constraint_name}").format(
            table=sql.Identifier(table_name),
            constraint_name=sql.Identifier(constraint_name))
        try:
            cursor_target.execute(query)
            connection_target.commit()
        except psycopg2.DatabaseError as error:
            sys.exit('Could not delete constraint.')


def create_database(connection, cursor: cursor, db_name: str, owner_name: str):
    print(f'Creating the "{db_name}" database...')

    cursor.execute(
        f"SELECT 1 FROM pg_catalog.pg_database WHERE datname = '{db_name}'")
    exists = cursor.fetchone()

    if not exists:
        try:
            if owner_name is not None:
                cursor.execute(sql.SQL("CREATE DATABASE {} OWNER {}").format(
                    sql.Identifier(db_name),
                    sql.Identifier(owner_name)))
            else:
                cursor.execute(sql.SQL("CREATE DATABASE {}").format(
                    sql.Identifier(db_name)))

            connection.commit()
        except psycopg2.DatabaseError as error:
            sys.exit('Database "{0}" could not be created. Error: {1}'.format(
                db_name, error))
    else:
        sys.exit(
            f'The database you tried to create "{db_name}" already exists. Please specify a new database name.')


def analyze_database(cursor: cursor, db_name: str) -> None:
    try:
        print('Retrieving statistics from the "{0}" database.'.format(db_name))
        cursor.execute("VACUUM ANALYZE;")
    except psycopg2.DatabaseError as error:
        sys.exit('Database "{0}" could not be analyzed. Error: {1}'.format(
            db_name, error))


def show_database_stats(cursor: cursor, tables_arg: str):
    tables = get_tables(cursor)

    tables_list = None
    if tables_arg:
        tables_list = tables_arg.split(",")
        tables_list = [table.strip(' ') for table in tables_list]

    for table_info in tables:
        table_name = table_info[1]

        if tables_list:
            if table_name not in tables_list:
                continue
        sub_result = get_table_stats(cursor, table_name)

        print(f'\n -- {table_name} -- \n')
        print(sub_result)


def get_tables(cursor: cursor):
    try:
        cursor.execute("""
        SELECT 
            nspname AS schemaname,relname as tablename, reltuples::bigint as rowcount,rank() over(order by reltuples desc)
        FROM pg_class C
        LEFT JOIN pg_namespace N ON (N.oid = C.relnamespace)
        WHERE 
            nspname NOT IN ('pg_catalog', 'information_schema') AND
            relkind='r' 
        ORDER BY tablename;""")

        return cursor.fetchall()
    except psycopg2.DatabaseError as error:
        sys.exit(
            'Could not retrieve the database\'s table information. Error description: {0}'.format(error))


def get_table_stats(cursor: cursor, table_name: str):
    try:
        cursor.execute(f"""
           select 
            attname, null_frac, avg_width, n_distinct, 
            most_common_vals, most_common_freqs, histogram_bounds, 
            correlation 
           from pg_stats 
          where schemaname not in ('pg_catalog') 
            and tablename = '{table_name}'
            and inherited = false;""")

        return cursor.fetchall()
    except psycopg2.DatabaseError as error:
        sys.exit('Could not get statistics for the "{0}" table. Error description: {1}'.format(
            table_name, error))


def get_column_information(cursor: cursor, table_name: str):
    try:
        cursor.execute(f"""
            SELECT 
                column_name, data_type, character_maximum_length,
                column_default,
                numeric_precision, numeric_precision_radix, numeric_scale
            FROM   information_schema.columns
            WHERE  table_name = '{table_name}'
            ORDER  BY ordinal_position;
            """)

        return cursor.fetchall()
    except psycopg2.DatabaseError as error:
        sys.exit('Could not get columns for the "{0}" table. Error description: {1}'.format(
            table_name, error))


def get_constr_information(cursor: cursor):
    try:
        cursor.execute(f"""
                    SELECT kcu.column_name AS foreign_column,
                            tc.table_name AS foreign_table,   
                            ccu.column_name "referenced primary_column",
                            ccu.table_name AS "referenced primary_table",
                            tc.constraint_type,
                            kcu.constraint_name		
                      FROM information_schema.table_constraints AS tc 
                      JOIN information_schema.key_column_usage AS kcu
                        ON tc.constraint_name = kcu.constraint_name
                       AND tc.table_schema = kcu.table_schema
                      JOIN information_schema.constraint_column_usage AS ccu
                        ON ccu.constraint_name = tc.constraint_name
                       AND ccu.table_schema = tc.table_schema;""")

        return cursor.fetchall()
    except psycopg2.DatabaseError as error:
        sys.exit('Could not constraint information')




from src import data_generator

# Geometry types


def test_generate_no_constraints_point():
    data_type = 'point'
    rows_to_gen = 20
    avg_width = None
    table_name = 'geom_types'
    min_bound = 1
    max_bound = 21
    numeric_precision = 32

    values = data_generator.generate_no_constraint(data_type=data_type,
                                                   rows_to_gen=rows_to_gen,
                                                   avg_width=avg_width,
                                                   table_name=table_name,
                                                   numeric_precision=numeric_precision,
                                                   min_bound=min_bound,
                                                   max_bound=max_bound)

    max_value = max(values)
    min_value = min(values)

    assert values
    assert len(values) == 20
    #assert values[0] == '(0,0)'


def test_generate_no_constraints_line():
    data_type = 'line'
    rows_to_gen = 20
    avg_width = None
    table_name = 'geom_types'
    min_bound = 1
    max_bound = 21
    numeric_precision = 32

    values = data_generator.generate_no_constraint(data_type=data_type,
                                                   rows_to_gen=rows_to_gen,
                                                   avg_width=avg_width,
                                                   table_name=table_name,
                                                   numeric_precision=numeric_precision,
                                                   min_bound=min_bound,
                                                   max_bound=max_bound)

    max_value = max(values)
    min_value = min(values)

    assert values
    assert len(values) == 20
    #assert values[0] == '{1,2,3}'


def test_generate_no_constraints_polygon():
    data_type = 'polygon'
    rows_to_gen = 20
    avg_width = None
    table_name = 'geom_types'
    min_bound = 1
    max_bound = 21
    numeric_precision = 32

    values = data_generator.generate_no_constraint(data_type=data_type,
                                                   rows_to_gen=rows_to_gen,
                                                   avg_width=avg_width,
                                                   table_name=table_name,
                                                   numeric_precision=numeric_precision,
                                                   min_bound=min_bound,
                                                   max_bound=max_bound)

    max_value = max(values)
    min_value = min(values)

    assert values
    assert len(values) == 20
    #assert values[0] == '((0,0),(1,1),(0,0))'


def test_generate_no_constraints_circle():
    data_type = 'circle'
    rows_to_gen = 20
    avg_width = None
    table_name = 'geom_types'
    min_bound = 1
    max_bound = 21
    numeric_precision = 32

    values = data_generator.generate_no_constraint(data_type=data_type,
                                                   rows_to_gen=rows_to_gen,
                                                   avg_width=avg_width,
                                                   table_name=table_name,
                                                   numeric_precision=numeric_precision,
                                                   min_bound=min_bound,
                                                   max_bound=max_bound)

    max_value = max(values)
    min_value = min(values)

    assert values
    assert len(values) == 20
    #assert values[0] == '<(0,0),1>'


def test_generate_no_constraints_box():
    data_type = 'box'
    rows_to_gen = 20
    avg_width = None
    table_name = 'geom_types'
    min_bound = 1
    max_bound = 21
    numeric_precision = 32

    values = data_generator.generate_no_constraint(data_type=data_type,
                                                   rows_to_gen=rows_to_gen,
                                                   avg_width=avg_width,
                                                   table_name=table_name,
                                                   numeric_precision=numeric_precision,
                                                   min_bound=min_bound,
                                                   max_bound=max_bound)

    max_value = max(values)
    min_value = min(values)

    assert values
    assert len(values) == 20
    #assert values[0] == '((0,1),(1,0))'


def test_generate_no_constraints_lseg():
    data_type = 'lseg'
    rows_to_gen = 20
    avg_width = None
    table_name = 'geom_types'
    min_bound = 1
    max_bound = 21
    numeric_precision = 32

    values = data_generator.generate_no_constraint(data_type=data_type,
                                                   rows_to_gen=rows_to_gen,
                                                   avg_width=avg_width,
                                                   table_name=table_name,
                                                   numeric_precision=numeric_precision,
                                                   min_bound=min_bound,
                                                   max_bound=max_bound)

    max_value = max(values)
    min_value = min(values)

    assert values
    assert len(values) == 20
    #assert values[0] == '((0,1),(2,2))'


# binary types
def test_generate_no_constraints_bytea():
    data_type = 'bytea'
    rows_to_gen = 20
    avg_width = None
    table_name = 'geom_types'
    min_bound = 1
    max_bound = 21
    numeric_precision = 32

    values = data_generator.generate_no_constraint(data_type=data_type,
                                                   rows_to_gen=rows_to_gen,
                                                   avg_width=avg_width,
                                                   table_name=table_name,
                                                   numeric_precision=numeric_precision,
                                                   min_bound=min_bound,
                                                   max_bound=max_bound)

    max_value = max(values)
    min_value = min(values)

    print(values)

    assert values
    assert len(values) == 20



import numpy as np
from typing import List, Any
from src.generator import generator_utils


data_type_dict = {
    'smallint': int,
    'integer': int,
    'bigint': int,
    'decimal': int,
    'real': float,
    'double precision': float,
    'numeric': float
}


def get_numeric_vals_no_constraints(data_type, rows_to_gen, min_bound, max_bound,
                                    numeric_precision, numeric_precision_radix,
                                    numeric_scale, histogram):
    generated_vals = list()
    if generator_utils.fulfill_histogram_properties(histogram, rows_to_gen, data_type_dict[data_type]):

        while len(generated_vals) < rows_to_gen:
            # pick random number from buckets since all have same size
            random_pick = np.random.randint(0, len(histogram)-1)

            # get local bound min + max
            local_min = data_type_dict[data_type](histogram[random_pick])
            local_max = data_type_dict[data_type](histogram[random_pick+1])
            try:

                partial_generated_vals = generator_utils.random_number(numeric_precision,
                                                                       numeric_precision_radix,
                                                                       numeric_scale, min_value=local_min,
                                                                       max_value=local_max)

                generated_vals.append(partial_generated_vals)

            except ValueError:
                print('Sample size exceeded population size.')
    else:
        for _ in range(rows_to_gen):
            generated_vals.append(
                generator_utils.random_number(numeric_precision,
                                              numeric_precision_radix,
                                              numeric_scale,
                                              min_value=min_bound, max_value=max_bound))

    return generated_vals


def get_numeric_vals_foreign_constraint(data_type, rows_to_gen: int,
                                        pk_values, ref_table, ref_column,
                                        min_bound: Any = None,
                                        max_bound: Any = None,
                                        histogram: List[any] = None):

    error_count = 0
    generated_vals = list()
    pk_values_list = pk_values[ref_table][ref_column]

    if generator_utils.fulfill_histogram_properties(histogram, rows_to_gen,
                                                    data_type_dict[data_type]):
        # sort pks such we can use bisect
        pk_values_list.sort()
        while len(generated_vals) < rows_to_gen:
            # pick random number from histogram buckets
            random_pick = np.random.randint(0, len(histogram)-1)
            # get local bound min + max
            local_min = int(histogram[random_pick])
            local_max = int(histogram[random_pick+1])
            try:
                # check if local histogram bounds of original attribute is in pk.
                # if not take the closest value from pk_list
                if local_min in pk_values_list:
                    min_index = pk_values_list.index(local_min)
                else:
                    min_index = pk_values_list.index(
                        generator_utils.take_closest(pk_values_list, local_min))
                if local_max in pk_values_list:
                    max_index = pk_values_list.index(local_max)
                else:
                    max_index = pk_values_list.index(
                        generator_utils.take_closest(pk_values_list, local_max))
                pk_values_boundaries_list = pk_values_list[min_index:max_index]
                partial_generated_vals = generator_utils.random.choice(
                    pk_values_boundaries_list)
                generated_vals.append(partial_generated_vals)
            except ValueError:
                error_count += 1
    else:
        counter = 0
        if rows_to_gen <= len(pk_values_list):
            while len(generated_vals) < rows_to_gen:
                for value in pk_values_list:
                    if len(generated_vals) < rows_to_gen:
                        if counter < len(pk_values_list):
                            counter = counter + 1
                            if (min_bound <= value <= max_bound
                                    and value in pk_values_list):
                                generated_vals.append(value)
                        else:
                            l_min_bound = min(pk_values_list)
                            l_max_bound = max(pk_values_list)
                            if (l_min_bound <= value <= l_max_bound
                                    and value in pk_values_list):
                                generated_vals.append(value)

    return generated_vals


def get_numeric_vals_primary_constraint(data_type, rows_to_gen: int, min_value: Any = None,
                                        max_value: Any = None, histogram: List[any] = None):

    generated_vals = list()

    if generator_utils.fulfill_histogram_properties(histogram, rows_to_gen,
                                                    data_type_dict[data_type]):
        while len(generated_vals) < rows_to_gen:
            # pick random number from buckets since all have same size
            random_pick = np.random.randint(0, len(histogram)-1)
            # get local bound min + max
            local_min = int(histogram[random_pick])
            local_max = int(histogram[random_pick+1])
            try:
                # get random number from of the local bounds
                partial_generated_vals = np.random.randint(
                    local_min, local_max)
                if partial_generated_vals not in generated_vals:
                    generated_vals.append(partial_generated_vals)
            except ValueError:
                print('Sample size exceeded population size.')
    else:
        if rows_to_gen > (max_value - min_value):
            max_value = min_value + rows_to_gen
        try:
            generated_vals = generator_utils.random.sample(
                range(min_value, max_value), int(rows_to_gen))
        except ValueError:
            print('Sample size exceeded population size.')
    return generated_vals



from typing import List
from src.generator import generator_utils


def get_varchar_no_constraints(rows_to_gen: int, avg_width: int, list_for_bound: List[str] = None):
    generated_vals = list()

    if list_for_bound is not None:
        value = list_for_bound[0]
    else:
        value = None

    for _ in range(rows_to_gen):
        generated_vals.append(generator_utils.random_word(
            avg_width,
            value=value))

    return generated_vals


def get_varchar_foreign_constraint(rows_to_gen: int, pk_values, ref_table, ref_column):
    generated_vals = list()

    pk_values_list = pk_values[ref_table][ref_column]
    while len(generated_vals) < rows_to_gen:
        for value in pk_values_list:
            if len(generated_vals) < rows_to_gen:
                generated_vals.append(value)

    return generated_vals


def get_varchar_primary_constraints(rows_to_gen: int, avg_width: int, list_for_bound: List[str] = None):
    generated_vals = list()
    for _ in range(rows_to_gen):
        value = generator_utils.random_word(avg_width - 1,
                                            value=list_for_bound[generator_utils.gen_random_number(0, len(list_for_bound) - 1)])
        if value not in generated_vals:
            generated_vals.append(value)

    return generated_vals



from bisect import bisect_left
from typing import List, Optional, Sequence, Union
import string
import random
from typing import Any
import radar


def fulfill_histogram_properties(histogram, rows_to_gen, data_type=float):
    if histogram is None:
        return False
        
    max_nr_of_rows_from_bounds = data_type(
        histogram[len(histogram)-1])-data_type(histogram[0]) + 1

    if max_nr_of_rows_from_bounds < rows_to_gen:
        return True

    return False


def take_closest(myList: List[any], myNumber: int):
    pos = bisect_left(myList, myNumber)
    if pos == 0:
        return myList[0]
    if pos == len(myList):
        return myList[-1]
    before = myList[pos - 1]
    after = myList[pos]
    if after - myNumber < myNumber - before:
        return after
    else:
        return before


def random_word(average_length: int, value: str = None) -> str:
    average_length = round(average_length-1)
    if value:
        if str(value).isdigit():
            word = gen_random_word(average_length, numeric=True)
        elif str(value).isupper():
            word = gen_random_word(average_length).upper()
        elif str(value) and str(value)[0].isupper():
            word = gen_random_word(average_length).capitalize()
        else:
            word = gen_random_word(average_length)
    else:
        word = gen_random_word(average_length)

    return word


def gen_random_word(length: int, numeric: bool = False) -> str:
    if length == 0:
        length += 1
    letters = string.ascii_lowercase
    numbers = string.digits

    if numeric:
        return ''.join(random.choice(numbers) for i in range(length))
    else:
        return ''.join(random.choice(letters) for i in range(length))


def gen_random_number(start: Any, end: Any, uniform: bool = False) -> Union[int, float]:

    start = int(start)
    end = int(end)

    if uniform:
        return random.uniform(start, end)
    else:
        return random.randint(start, end)


def random_number(numeric_precision: int = None, numeric_precision_radix: int = None, numeric_scale: int = None,
                  min_value: Any = None, max_value: Any = None) -> Union[int, float]:

    if numeric_precision:
        if numeric_scale and numeric_scale != 0:
            number = round(gen_random_number(
                min_value or 0,
                max_value or ((numeric_precision_radix **
                              (numeric_precision - numeric_scale - 1)) / 1.5),
                uniform=True),
                numeric_scale)
        else:
            number = gen_random_number(min_value or 0,
                                       max_value or ((numeric_precision_radix ** (numeric_precision - 1)) / 1.5))
    else:
        number = gen_random_number(0, 50000)

    return number


def random_date(start_date, end_date, time=False):
    if time:
        return radar.random_datetime(start=start_date, stop=end_date)
    else:
        return radar.random_date(start=start_date, stop=end_date)


def random_boolean(postgres: bool = True) -> Union[str, bool]:

    random_boolean = random.choice([True, False])
    if not postgres:
        return random_boolean
    else:
        if random_boolean:
            return 'true'
        else:
            return 'false'


def random_choices(list: List[Any], weights: Optional[Sequence[float]], k: int = 1) -> Any:
    if k == 1:
        return random.choices(list, weights=weights, k=k)[0]
    else:
        return random.choices(list, weights=weights, k=k)
